/* Copyright 2009 Mel Wilson
This file is part of SysvIpcMsg.

SysvIpcMsg is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

SysvIpcMsg is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with SysvIpcMsg.  If not, see <http://www.gnu.org/licenses/>.
*/
/* $Id$ */

#include <Python.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>

/*=======================================================*/
/*
  Utility
*/

static PyObject *sysvipc_ftok (PyObject *self, PyObject *args)
// In Python, call with
//	string path name
//	int project id
// returns
//	int id number.
{
	char *pathname;
	int proj_id;
	key_t idnum;
	if (!PyArg_ParseTuple(args, "si", &pathname, &proj_id))
		return NULL;

	idnum = ftok (pathname, proj_id);
	return Py_BuildValue ("i", (long)idnum);
} /* sysvipc_msgget */


/*=======================================================*/
/*
  Routines for message handling
*/

/* Set up a message buffer structure
 * big enough for a message being sent
 * or to be received.
 */
static struct msgbuf *get_msgbuf (const char *mtext, size_t textsize, long mtype)
{
	struct msgbuf *p = malloc (sizeof (struct msgbuf) + textsize);
	if (!p)
		return (struct msgbuf *)PyErr_NoMemory();
		
	p->mtype = mtype;
	if (mtext)
		memcpy ((char *)&(p->mtext), mtext, textsize);
	else
		strcpy ((char *)&(p->mtext), "");
	return p;
} /* get_msgbuf */


static PyObject *sysvipc_msgget (PyObject *self, PyObject *args)
// In Python, call with
//	int message key value
//	int message flags
// returns
//	int message queue id.
{
	int qid;
	int i_key, msgflg;
	if (!PyArg_ParseTuple(args, "ii", &i_key, &msgflg))
		return NULL;

	qid = msgget ((key_t)i_key, msgflg);
	if (qid < 0) {
		PyErr_SetFromErrno (PyExc_IOError);
		return NULL;
	}
	return Py_BuildValue ("i", qid);
} /* sysvipc_msgget */


static PyObject *sysvipc_msgctl_set (PyObject *self, PyObject *args)
// In Python, call with
//	int message_queue_id
//	dict containing msqid_ds fields to be set, indexed by field name as string, 
// returns
//	int result of msgctl (..., IPC_SET, ...);
{
	int msqid, result, ppos;
	struct msqid_ds buf;
	const char *s;
	int i;
	PyObject *d=NULL;
	PyObject *k=NULL, *v=NULL;
	
	if (!PyArg_ParseTuple(args, "iO", &msqid, &d))
		return NULL;
		
	if (!PyDict_Check (d)) {
		PyErr_SetString (PyExc_TypeError, "Argument 2 must be a dict.");
		return NULL;
	}
		
	result = msgctl (msqid, IPC_STAT, &buf);
	if (result < 0) {
		PyErr_SetFromErrno (PyExc_IOError);
		return NULL;
	}
	
	ppos = 0;
	while (PyDict_Next (d, &ppos, &k, &v)) {
		if (!PyString_Check (k))
			continue;
		if (!PyInt_Check (v))
			continue;
		s = PyString_AsString (k);
		i = PyInt_AS_LONG (v);
		
		if (0 == strcmp (s, "uid")) {
			buf.msg_perm.uid = (__uid_t)i;
		}
		else if (0 == strcmp (s, "gid")) {
			buf.msg_perm.gid = (__gid_t)i;
		}
		else if (0 == strcmp (s, "mode")) {
			buf.msg_perm.mode = (unsigned short int)i;
		}
		else if (0 == strcmp (s, "msg_qbytes")) {
			buf.msg_qbytes = (msglen_t)i;
		}
	} /* while */
		
	result = msgctl (msqid, IPC_SET, &buf);
	if (result < 0) {
		PyErr_SetFromErrno (PyExc_IOError);
		return NULL;
	}
	
	return Py_BuildValue ("i", result);	
} /* sysvipc_msgctl_set */


static PyObject *sysvipc_msgctl_stat (PyObject *self, PyObject *args)
// In Python, call with
//	int message_queue_id
// returns
//	dict containing fields from msqid_ds, indexed by field name as string
{
	int msqid, result;
	struct msqid_ds buf;
	PyObject *d = NULL;

	if (!PyArg_ParseTuple(args, "i", &msqid))
		return NULL;

	result = msgctl (msqid, IPC_STAT, &buf);
	if (result < 0) {
		PyErr_SetFromErrno (PyExc_IOError);
		return NULL;
	}

	d = PyDict_New();
	if (!d)
		return NULL;

	PyDict_SetItemString (d, "__key", PyInt_FromLong ((long)buf.msg_perm.__key));
	PyDict_SetItemString (d, "uid", PyInt_FromLong ((long)buf.msg_perm.uid));
	PyDict_SetItemString (d, "gid", PyInt_FromLong ((long)buf.msg_perm.gid));
	PyDict_SetItemString (d, "cuid", PyInt_FromLong ((long)buf.msg_perm.cuid));
	PyDict_SetItemString (d, "cgid", PyInt_FromLong ((long)buf.msg_perm.cgid));
	PyDict_SetItemString (d, "mode", PyInt_FromLong ((long)buf.msg_perm.mode));
	PyDict_SetItemString (d, "__seq", PyInt_FromLong ((long)buf.msg_perm.__seq));
	
	PyDict_SetItemString (d, "msg_stime", PyInt_FromLong ((long)buf.msg_stime));
	PyDict_SetItemString (d, "msg_rtime", PyInt_FromLong ((long)buf.msg_rtime));
	PyDict_SetItemString (d, "msg_ctime", PyInt_FromLong ((long)buf.msg_ctime));
	PyDict_SetItemString (d, "__msg_cbytes", PyInt_FromLong ((long)buf.__msg_cbytes));
	PyDict_SetItemString (d, "msg_qnum", PyInt_FromLong ((long)buf.msg_qnum));
	PyDict_SetItemString (d, "msg_qbytes", PyInt_FromLong ((long)buf.msg_qbytes));
	PyDict_SetItemString (d, "msg_lspid", PyInt_FromLong ((long)buf.msg_lspid));
	PyDict_SetItemString (d, "msg_lrpid", PyInt_FromLong ((long)buf.msg_lrpid));

	return d;	
	
} /* sysvipc_msgctl_stat */


static PyObject *sysvipc_msgctl_rmid (PyObject *self, PyObject *args)
// In Python, call with
//	int message_queue_id
// returns
//	int result of msgctl (..., IPC_RMID, ...);
{
	int msqid, result;
	struct msqid_ds buf ;
	if (!PyArg_ParseTuple(args, "i", &msqid))
		return NULL;
	
	result = msgctl (msqid, IPC_RMID, &buf);
	if (result < 0) {
		PyErr_SetFromErrno (PyExc_IOError);
		return NULL;
	}
	
	return Py_BuildValue ("i", result);	
} /* sysvipc_msgctl_rmid */


static PyObject *sysvipc_msgrcv (PyObject *self, PyObject *args)
// In Python, call with
//	int message_queue_id
//	int length limit (in characters) of the message to receive
//	int type-code of the message to receive
//	int message flags
// returns
//	string containing received message, empty string if no message received
{
	int msqid, msgflg;
	long msgsz, msgtyp;
	struct msgbuf *msgp;
	ssize_t rcvsize;
	PyObject *result;
	
	if (!PyArg_ParseTuple(args, "illi", &msqid, &msgsz, &msgtyp, &msgflg))
		return NULL;
	
	msgp = get_msgbuf (NULL, msgsz, msgtyp);
	if (!msgp)
		return NULL;
	
	rcvsize = msgrcv (msqid, msgp, (size_t)msgsz, msgtyp, msgflg);
	if (rcvsize < 0) {
		PyErr_SetFromErrno (PyExc_IOError);
		free (msgp);
		return NULL;
	}
	
	result = Py_BuildValue("s#", &(msgp->mtext), rcvsize);
	free (msgp);
	return result;
} /* sysvipc_msgrcv */


static PyObject *sysvipc_msgsnd (PyObject *self, PyObject *args)
// In Python, call with
//	int message_queue_id
//  	string containing text of message to send
//	int type-code of the message to send
//	int message flags
// returns
//	int result of msgsnd ( ... );
{
	int msqid, msgflg, result;
	long msgtyp;
	const char *text = NULL;
	struct msgbuf *msgp;
	size_t msgsz;
	
	if (!PyArg_ParseTuple(args, "is#li", &msqid, &text, &msgsz, &msgtyp, &msgflg))
		return NULL;
	
	msgp = get_msgbuf (text, msgsz, msgtyp);
	if (!msgp)
		return NULL;
	
	result = msgsnd (msqid, msgp, (size_t)msgsz, msgflg);	
	if (result < 0) {
		PyErr_SetFromErrno (PyExc_IOError);
		free (msgp);
		return NULL;
	}
	free (msgp);
	return Py_BuildValue ("i", (int)result);
} /* sysvipc_msgsnd */


/*=======================================================*/

static PyMethodDef SysvipcmsgMethods[] = {
	{"ftok", sysvipc_ftok, METH_VARARGS, "Convert a pathname and a project identifier to a System V IPC key."},

	{"msgget", sysvipc_msgget, METH_VARARGS, "Get an IPC Message queue ID."},
	{"msgctl_set", sysvipc_msgctl_set, METH_VARARGS, "Set status information for a message queue."},
	{"msgctl_stat", sysvipc_msgctl_stat, METH_VARARGS, "Get status information for a message queue."},
	{"msgctl_rmid", sysvipc_msgctl_rmid, METH_VARARGS, "Remove a message queue."},
	{"msgrcv", sysvipc_msgrcv, METH_VARARGS, "Receive a message from an IPC Message queue."},
	{"msgsnd", sysvipc_msgsnd, METH_VARARGS, "Send a message on an IPC Message queue."},

	{NULL, NULL, (int)NULL, NULL}
};

PyMODINIT_FUNC init_sysvipcmsg (void)
{
	PyObject *module;
	module = Py_InitModule ("_sysvipcmsg", SysvipcmsgMethods);

	// flag bits for msgget ...
	PyModule_AddIntConstant (module, "IPC_CREAT", IPC_CREAT);
	PyModule_AddIntConstant (module, "IPC_EXCL", IPC_EXCL);

	// special key value for msgget ...
	PyModule_AddIntConstant (module, "IPC_PRIVATE", (int)IPC_PRIVATE);

	// Permission bits for msgget flags ...
	PyModule_AddIntConstant (module, "S_IRUSR", S_IRUSR);	// user read
	PyModule_AddIntConstant (module, "S_IWUSR", S_IWUSR);	// user write
	PyModule_AddIntConstant (module, "S_IREAD", S_IREAD);	// same as user read
	PyModule_AddIntConstant (module, "S_IWRITE", S_IWRITE);	// same as user write
	PyModule_AddIntConstant (module, "S_IRGRP", S_IRGRP);	// group read
	PyModule_AddIntConstant (module, "S_IWGRP", S_IWGRP);	// group write
	PyModule_AddIntConstant (module, "S_IROTH", S_IROTH);	// others read
	PyModule_AddIntConstant (module, "S_IWOTH", S_IWOTH);	// others write

	// flags for msgrcv, msgsnd ...
	PyModule_AddIntConstant (module, "IPC_INFO", IPC_INFO);
	PyModule_AddIntConstant (module, "IPC_NOWAIT", IPC_NOWAIT);

	// msgctl commands ...
	PyModule_AddIntConstant (module, "IPC_RMID", IPC_RMID);
	PyModule_AddIntConstant (module, "IPC_SET", IPC_SET);
	PyModule_AddIntConstant (module, "IPC_STAT", IPC_STAT);

	// flag bits for msgrcv, msgsnd ...
	PyModule_AddIntConstant (module, "MSG_EXCEPT", MSG_EXCEPT);
	PyModule_AddIntConstant (module, "MSG_INFO", MSG_INFO);
	PyModule_AddIntConstant (module, "MSG_NOERROR", MSG_NOERROR);
	PyModule_AddIntConstant (module, "MSG_STAT", MSG_STAT);
}
